from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, TypeVar

from typing_extensions import Concatenate, ParamSpec

from .sign import FnImplement, FnRecord

if TYPE_CHECKING:
    from .collector import BaseCollector
    from .fn import Fn
    from .perform import BasePerform
    from .staff import Staff

T = TypeVar("T")
N = TypeVar("N", bound="BasePerform")
R = TypeVar("R", covariant=True)
P = ParamSpec("P")


class OverloadBehavior:
    def harvest_record(self, staff: Staff, fn: Fn) -> FnRecord:
        result = staff.artifact_map.get(FnImplement(fn))
        if result is None:
            raise NotImplementedError
        return result

    def harvest_overload(
        self, staff: Staff, fn: Fn[P, R], *args: P.args, **kwargs: P.kwargs
    ) -> tuple[BaseCollector, Callable[Concatenate[Any, P], R]]:
        artifact_record = self.harvest_record(staff, fn)

        if fn.has_overload_capability:
            bound_args = fn.shape_signature.bind(*args, **kwargs)
            bound_args.apply_defaults()
            collections = None

            for overload_item, required_args in fn.overload_param_map.items():
                scope = artifact_record["overload_scopes"][overload_item.identity]
                entities = overload_item.get_entities(scope, {i: bound_args.arguments[i] for i in required_args})
                collections = entities if collections is None else collections.intersection(entities)

            if not collections:
                raise NotImplementedError

            return collections.pop()  # type: ignore

        else:
            return artifact_record["record_tuple"]  # type: ignore



DEFAULT_BEHAVIOR = OverloadBehavior()



from avilla.core.ryanvk.collector.context import (
    ContextBasedPerformTemplate as ContextBasedPerformTemplate,
)
from avilla.core.ryanvk.collector.context import ContextCollector as ContextCollector
from avilla.core.ryanvk.descriptor.query import QueryRecord as QueryRecord
from avilla.core.ryanvk.descriptor.query import QuerySchema as QuerySchema
from avilla.core.ryanvk.overload.metadata import MetadataOverload as MetadataOverload
from avilla.core.ryanvk.overload.target import TargetOverload as TargetOverload
from graia.ryanvk.capability import Capability as Capability
from graia.ryanvk.fn import Fn as Fn
from graia.ryanvk.overload import FnOverload as FnOverload
from graia.ryanvk.overload import SimpleOverload as SimpleOverload
from graia.ryanvk.overload import TypeOverload as TypeOverload



from graia.amnesia.message import MessageChain as MessageChain

from avilla.core.account import BaseAccount as BaseAccount
from avilla.core.application import Avilla as Avilla
from avilla.core.builtins.capability import CoreCapability as CoreCapability
from avilla.core.context import Context as Context
from avilla.core.dispatchers import AvillaBuiltinDispatcher as AvillaBuiltinDispatcher
from avilla.core.elements import Audio as Audio
from avilla.core.elements import File as File
from avilla.core.elements import Notice as Notice
from avilla.core.elements import NoticeAll as NoticeAll
from avilla.core.elements import Picture as Picture
from avilla.core.elements import Text as Text  # noqa: F401
from avilla.core.elements import Unknown as Unknown
from avilla.core.elements import Video as Video
from avilla.core.event import AvillaEvent as AvillaEvent
from avilla.core.event import DirectSessionCreated as DirectSessionCreated
from avilla.core.event import DirectSessionDestroyed as DirectSessionDestroyed
from avilla.core.event import MemberCreated as MemberCreated
from avilla.core.event import MemberDestroyed as MemberDestroyed
from avilla.core.event import MetadataModified as MetadataModified
from avilla.core.event import RelationshipCreated as RelationshipCreated
from avilla.core.event import RelationshipDestroyed as RelationshipDestroyed
from avilla.core.event import RelationshipEvent as RelationshipEvent
from avilla.core.event import SceneCreated as SceneCreated
from avilla.core.event import SceneDestroyed as SceneDestroyed
from avilla.core.exceptions import AccountDeleted as AccountDeleted
from avilla.core.exceptions import AccountMuted as AccountMuted
from avilla.core.exceptions import ActionFailed as ActionFailed
from avilla.core.exceptions import ContextError as ContextError
from avilla.core.exceptions import DeprecatedError as DeprecatedError
from avilla.core.exceptions import HttpRequestError as HttpRequestError
from avilla.core.exceptions import InaccessibleInterface as InaccessibleInterface
from avilla.core.exceptions import InvalidAuthentication as InvalidAuthentication
from avilla.core.exceptions import InvalidOperation as InvalidOperation
from avilla.core.exceptions import NetworkError as NetworkError
from avilla.core.exceptions import ParserException as ParserException
from avilla.core.exceptions import RemoteError as RemoteError
from avilla.core.exceptions import TooLongMessage as TooLongMessage
from avilla.core.exceptions import UnknownError as UnknownError
from avilla.core.exceptions import UnknownTarget as UnknownTarget
from avilla.core.exceptions import UnsupportedOperation as UnsupportedOperation
from avilla.core.message import Message as Message
from avilla.core.metadata import Metadata as Metadata
from avilla.core.platform import Abstract as Abstract
from avilla.core.platform import Branch as Branch
from avilla.core.platform import Land as Land
from avilla.core.platform import Maintainer as Maintainer
from avilla.core.platform import Platform as Platform
from avilla.core.platform import PlatformDescription as PlatformDescription
from avilla.core.platform import Version as Version
from avilla.core.protocol import BaseProtocol as BaseProtocol
from avilla.core.resource import LocalFileResource as LocalFileResource
from avilla.core.resource import RawResource as RawResource
from avilla.core.resource import Resource as Resource
from avilla.core.resource import UrlResource as UrlResource
from avilla.core.selector import Selectable as Selectable
from avilla.core.selector import Selector as Selector
from avilla.core.service import AvillaService as AvillaService
from avilla.core.typing import Ensureable as Ensureable
from avilla.standard.core.account import AccountAvailable as AccountAvailable
from avilla.standard.core.account import AccountStatusChanged as AccountStatusChanged
from avilla.standard.core.account import AccountUnavailable as AccountUnavailable
from avilla.standard.core.activity import ActivityAvailable as ActivityAvailable
from avilla.standard.core.activity import ActivityEvent as ActivityEvent
from avilla.standard.core.activity import ActivityTrigged as ActivityTrigged
from avilla.standard.core.activity import ActivityUnavailable as ActivityUnavailable
from avilla.standard.core.application import ApplicationClosed as ApplicationClosed
from avilla.standard.core.application import ApplicationClosing as ApplicationClosing
from avilla.standard.core.application import (
    ApplicationPreparing as ApplicationPreparing,
)
from avilla.standard.core.application import ApplicationReady as ApplicationReady
from avilla.standard.core.application import (
    AvillaLifecycleEvent as AvillaLifecycleEvent,
)
from avilla.standard.core.common import Count as Count
from avilla.standard.core.message import MessageEdit as MessageEdit
from avilla.standard.core.message import MessageEdited as MessageEdited
from avilla.standard.core.message import MessageReceived as MessageReceived
from avilla.standard.core.message import MessageRevoke as MessageRevoke
from avilla.standard.core.message import MessageRevoked as MessageRevoked
from avilla.standard.core.message import MessageSend as MessageSend
from avilla.standard.core.message import MessageSent as MessageSent
from avilla.standard.core.privilege import BanCapability as BanCapability
from avilla.standard.core.privilege import BanInfo as BanInfo
from avilla.standard.core.privilege import MuteAllCapability as MuteAllCapability
from avilla.standard.core.privilege import MuteCapability as MuteCapability
from avilla.standard.core.privilege import MuteInfo as MuteInfo
from avilla.standard.core.privilege import Privilege as Privilege
from avilla.standard.core.privilege import PrivilegeCapability as PrivilegeCapability
from avilla.standard.core.profile import Nick as Nick
from avilla.standard.core.profile import NickCapability as NickCapability
from avilla.standard.core.profile import Summary as Summary
from avilla.standard.core.profile import SummaryCapability as SummaryCapability
from avilla.standard.core.relation import SceneCapability as SceneCapability
from avilla.standard.core.request import Answers as Answers
from avilla.standard.core.request import Comment as Comment
from avilla.standard.core.request import Questions as Questions
from avilla.standard.core.request import Reason as Reason
from avilla.standard.core.request import RequestAccepted as RequestAccepted
from avilla.standard.core.request import RequestCancelled as RequestCancelled
from avilla.standard.core.request import RequestCapability as RequestCapability
from avilla.standard.core.request import RequestEvent as RequestEvent
from avilla.standard.core.request import RequestIgnored as RequestIgnored
from avilla.standard.core.request import RequestReceived as RequestReceived
from avilla.standard.core.request import RequestRejected as RequestRejected
from avilla.standard.core.resource import FileRemoved as FileRemoved
from avilla.standard.core.resource import FileUploaded as FileUploaded
from avilla.standard.core.resource import ResourceAvailable as ResourceAvailable
from avilla.standard.core.resource import ResourceEvent as ResourceEvent
from avilla.standard.core.resource import ResourceUnavailable as ResourceUnavailable



from __future__ import annotations

import asyncio
from typing import Coroutine

TASKS: dict[int, asyncio.Task] = {}


def queue_task(coro: Coroutine):
    task = asyncio.create_task(coro)
    task_id = id(task)
    TASKS[task_id] = task
    task.add_done_callback(lambda x: TASKS.pop(task_id))



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Generic, TypeVar, overload

from typing_extensions import Concatenate, ParamSpec, Self

if TYPE_CHECKING:
    from .collector import BaseCollector
    from .fn import Fn
    from .perform import BasePerform
    from .typing import SupportsCollect

P = ParamSpec("P")
R = TypeVar("R", covariant=True)
P1 = ParamSpec("P1")
Q = TypeVar("Q", contravariant=True)

WrapperCallable = Callable[[Q], Q]
FnRecord = tuple["BaseCollector", Callable]


class OverridePerformEntity(Generic[P, R]):
    collector: BaseCollector
    fn: Fn[P, R]
    entity: Callable[Concatenate[Any, P], R]

    origin_perform: type[BasePerform]
    name: str

    def __init__(
        self,
        collector: BaseCollector,
        fn: Fn[P, R],
        entity: Callable[Concatenate[Any, P], R],
    ) -> None:
        self.collector = collector
        self.fn = fn
        self.entity = entity

    @classmethod
    def collect(
        cls,
        collector: BaseCollector,
        fn: SupportsCollect[P1, WrapperCallable[Callable[Concatenate[Any, P], R]]],
        *args: P1.args,
        **kwargs: P1.kwargs,
    ):
        def wrapper(entity: Callable[Concatenate[Any, P], R]) -> OverridePerformEntity[P, R]:
            fn.collect(collector, *args, **kwargs)(entity)
            return cls(collector, fn, entity)  # type: ignore

        return wrapper

    def __set_name__(self, owner: type[BasePerform], name: str):
        self.origin_perform = owner
        self.name = name

    @overload
    def __get__(self, instance: None, owner: type[BasePerform]) -> Self:
        ...

    @overload
    def __get__(self, instance: OverridePerformAgent, owner: type) -> Self:
        ...

    @overload
    def __get__(self, instance: BasePerform, owner: type[BasePerform]) -> OverridePerformAgent[P, R]:
        ...

    def __get__(self, instance: Any, owner: type):
        if instance is None:
            return self

        return OverridePerformAgent(self, instance)


class OverridePerformAgent(Generic[P, R]):
    perform_entity: OverridePerformEntity[P, R]
    instance: BasePerform

    def __init__(self, perform_entity: OverridePerformEntity[P, R], instance: BasePerform) -> None:
        self.perform_entity = perform_entity
        self.instance = instance

    def __call__(self, *args: P.args, **kwds: P.kwargs) -> R:
        return self.perform_entity.entity(self.instance, *args, **kwds)

    def super(self, *args: P.args, **kwds: P.kwargs) -> R:
        origin = self.instance.staff
        collections = origin.artifact_collections[1:]
        staff_type = type(origin)
        staff = staff_type(collections, origin.components)
        staff.instances.update(origin.instances)
        staff.exit_stack = origin.exit_stack
        return staff.call_fn(self.perform_entity.fn, *args, **kwds)
